import { walk } from 'jsr:@std/fs'
import { join, relative, toFileUrl } from 'jsr:@std/path'

export type Constraint = {
	kind: 'file:path' | 'file:label' | 'object:label'
	format: string
	fields: {
		[name: string]: RegExp[]
	}
}

function checkConstraint(value: string, constraint: Constraint): boolean {
	const fields = constraint.format.match(/\$\w+/g) ?? []
	const regexp = fields.map((format) => constraint.fields[format.slice(1)])
	const matcher = constraint.format.split(
		new RegExp(fields.join('|').replaceAll('$', '\\$')),
	)

	for (const _index in matcher) {
		const index = Number(_index)

		//check pattern
		if (!value.startsWith(matcher[index])) {
			return false
		}

		//consume pattern
		value = value.slice(matcher[index].length)

		//check regexp
		if (regexp.at(index)?.every((reg) => value.match(reg)?.index !== 0)) {
			return false
		}

		//consume regexp
		value = value.slice(
			Math.max(
				...regexp.at(index)?.map((reg) =>
					value.match(reg)?.[0]?.length ?? 0
				) ?? [0],
			),
		)
	}

	//verified if all string is consumed
	return value.length === 0
}

function logCheckConstraint(constraint: Constraint, isValid: boolean) {
	if (isValid) {
		console.log(
			`%c[utils:freecad]%c: valid %c${constraint.kind}`,
			'color: green; font-weight: bold',
			'',
			'color: goldenrod',
		)
	} else {
		console.error(
			`%c[utils:freecad]%c: invalid %c${constraint.kind}`,
			'color: red; font-weight: bold',
			'',
			'color: goldenrod',
		)
	}
}

export function checkConstraints(
	fcinfo: string,
	constraints: Constraint[],
	{ basePath = Deno.cwd(), filePath }: {
		basePath?: string
		filePath: string
	},
): boolean {
	for (const constraint of constraints) {
		if (constraint.kind === 'file:path') {
			const isValid = checkConstraint(
				relative(basePath, filePath).replaceAll('\\', '/'),
				constraint,
			)
			logCheckConstraint(constraint, isValid)
			if (!isValid) return false
			continue
		}

		if (constraint.kind === 'file:label') {
			const match = fcinfo.match(/\s*Label : (.*)/)
			const isValid = checkConstraint(match?.at(1) ?? '', constraint)
			logCheckConstraint(constraint, isValid)
			if (!isValid) return false
			continue
		}

		if (constraint.kind === 'object:label') {
			const matches = [...fcinfo.matchAll(/\s*Label : (.*)/g)]
			const isValid = matches.slice(1).every((match) =>
				checkConstraint(match[1], constraint)
			)
			logCheckConstraint(constraint, isValid)
			if (!isValid) return false
		}
	}

	return true
}

if (import.meta.main) {
	const [constraintFile, basePath] = Deno.args
	const { default: constraints } = await import(
		toFileUrl(join(Deno.cwd(), constraintFile)).href
	)
	let exitCode = 0

	for await (
		const file of walk(basePath, {
			match: [/.+\.fcinfo/],
			includeDirs: false,
		})
	) {
		console.log(
			`%c[utils:freecad]%c: checking name convention for %c"${file.path}"`,
			'color: royalblue; font-weight: bold',
			'',
			'text-decoration: underline',
		)
		const fcinfo = await Deno.readTextFile(file.path)
		const isValid = checkConstraints(fcinfo, constraints, {
			basePath,
			filePath: file.path,
		})
		if (!isValid) {
			console.error(
				`%c[utils:freecad]%c: error in name convention for %c"${file.path}"`,
				'color: red; font-weight: bold',
				'',
				'text-decoration: underline',
			)
			exitCode = 1
		} else {
			console.log(
				`%c[utils:freecad]%c: valid name convention for %c"${file.path}"`,
				'color: green; font-weight: bold',
				'',
				'text-decoration: underline',
			)
		}
	}
	Deno.exit(exitCode)
}
