export class Translator {
	/**
	 * Parse a raw text line into a parsed Vegetal Signal serial event.
	 * @param raw Serial decoded text line.
	 * @returns Vegetal Signal event.
	 *
	 * @example
	 * ```ts
	 * const rawString = 'NORMAL MODE 0'
	 * const { kind, name, value } = Translator.parse(rawString)
	 * ```
	 */
	static parse(raw: string) {
		for (const info of Translator.#infos) {
			if (!info.matcher.test(raw)) continue
			const value = info.parse(raw)
			return Translator.#infoParser({ name: info.name, value })
		}

		for (const header of Translator.#headers) {
			if (!header.matcher.test(raw)) continue
			const value = header.parse(raw)
			return Translator.#headerParser({ name: header.name, value })
		}

		for (const serial of Translator.#serials) {
			if (!serial.matcher.test(raw)) continue
			const value = serial.parse(raw)
			return Translator.#serialParser({ value })
		}
	}

	/**
	 * Vegetal Signal Translator transform stream.
	 * Transform serial string messages into parsed and structured Vegetal Signal Datas.
	 *
	 * @example
	 * ```ts
	 * //Get vegetal signal device via web serial API
	 * const [ port ] = await navigator.serial.getPorts()
	 *
	 * //Pipe decoder and translator stream
	 * const events = port.readable
	 * 	.pipeThrough(new TextDecoderStream())
	 * 	.pipeThrough(Translator.transform)
	 *
	 * //Process events
	 * for await (const event of events) {
	 * 	console.table({
	 * 		event.kind,
	 * 		event.name,
	 * 		event.value
	 * 	})
	 * }
	 * ```
	 */
	static get transform() {
		return new TransformStream<
			string,
			ReturnType<typeof Translator['parse']>
		>({
			start() {},
			transform(chunk, controller) {
				chunk
					.split('\n')
					.map(Translator.parse)
					.filter((data) => data !== undefined)
					.forEach((data) => controller.enqueue(data))
			},
		})
	}

	static readonly #infos = [
		{
			name: 'power' as const,
			matcher: /^POWER (UP|DOWN): \d+ ms$/,
			parse: (value) =>
				value.match(/(UP|DOWN): (\d+)/)!.slice(1, 3) as [
					string,
					string,
				],
		},
		{
			name: 'ping' as const,
			matcher: /^OK$/,
			parse: () => true,
		},
		{
			name: 'format' as const,
			matcher: /^FORMAT:\w+$/,
			parse: (value) => value.split(':')[1],
		},
		{
			name: 'frequency' as const,
			matcher: /^FREQ:\d+\.\d+$/,
			parse: (value) => Number(value.split(':')[1]),
		},
		{
			name: 'channel' as const,
			matcher: /^CHANNEL:\d+$/,
			parse: (value) => Number(value.split(':')[1]),
		},
		{
			name: 'serial-number',
			matcher: /^SN:\w+-\w+/,
			parse: (value) => value.split(':')[1],
		},
	] satisfies ChunkParser[]

	static readonly #headers = [
		{
			name: 'serial-number' as const,
			matcher: /.*(SN \S+)$/,
			parse: (value) => value.split(' ')[1],
		},
		{
			name: 'firmware' as const,
			matcher: /^FW V\S+$/,
			parse: (value) => value.split('FW V')[1],
		},
		{
			name: 'power-level' as const,
			matcher: /^POWER \d+ \d+ \d+ \d+$/,
			parse: (value) =>
				value.match(/\d+/g)!.map(Number) as [
					number,
					number,
					number,
					number,
				],
		},
		{
			name: 'reboot' as const,
			matcher: /^REBOOT \d+, \d+$/,
			parse: (value) =>
				value.match(/\d+/g)!.map(Number) as [number, number],
		},
		{
			name: 'mode' as const,
			matcher: /^\w+ MODE \d+$/,
			parse: (value) => value.split(' MODE ')[0].toLowerCase(),
		},
	] satisfies ChunkParser[]

	static readonly #serials = [
		{
			name: 'serial' as const,
			matcher: /\d+,-?\d+,-?\d+,-?\d+,-?\d+,-?\d+,-?\d+,-?\d+/,
			parse: (value) =>
				value.split(',').map(Number) as [number, Channels],
		},
	] satisfies ChunkParser[]

	static #headerParser<Name extends string, Value>(
		{ name, value }: { name: Name; value: Value },
	) {
		return {
			kind: 'header',
			name,
			value,
		} as const
	}

	static #infoParser<Name extends string, Value>(
		{ name, value }: { name: Name; value: Value },
	) {
		if (name === 'power') {
			const [mode, timeout] = value as [string, string]

			return {
				kind: 'info',
				name,
				value: [mode.toLowerCase(), Number(timeout)],
			}
		}

		if (name === 'ping') {
			return {
				kind: 'info',
				name,
				value: true,
			}
		}

		return {
			kind: 'info',
			name,
			value: (value as string).toLowerCase(),
		} as const
	}

	static #serialParser({ value }: { value: [number, Channels] }) {
		const [timestamp, ...channels] = value

		return {
			kind: 'serial',
			timestamp,
			channels,
		} as const
	}
}

type Channels = [number, number, number, number, number, number, number, number]

type ChunkParser<T = unknown> = {
	readonly name: string
	matcher: RegExp
	parse: (raw: string) => T
}

//Example datas
// HEADER
/*
$(NOISE)SN VS823Cx-E661C8D84B56BC38<LF>
FW V4.2.1<LF>
POWER 4863 4880 67 69<LF>
REBOOT 0, 0<LF>
NORMAL MODE 0<LF>
*/
// INFOS
/*
POWER UP: 7153 ms<LF>
POWER DOWN: 11365 ms<LF>
*/
//SERIAL
/*
timestamp,Ch1,Ch2,Ch3,Ch4,Ch5,Ch6,Ch7,Ch8<LF>
*/
